import pydantic
import typing


class RuleMetadata(pydantic.BaseModel):
    name: typing.Optional[str] = pydantic.Field(
        None, description="The name of the rule"
    )
    version: typing.Optional[str] = pydantic.Field(
        ..., description="The version of the rule"
    )


class ExecutionSchema(pydantic.BaseModel):
    payload: typing.Optional[typing.Any] = pydantic.Field(
        None, description="The payload of the execution"
    )
    states: typing.Optional[typing.Any] = pydantic.Field(
        None, description="The states of the execution"
    )
    filters: typing.Optional[typing.Dict[str, typing.Any]] = pydantic.Field(
        None, description="The filters of the execution"
    )
    result: typing.Optional[typing.Any] = pydantic.Field(
        None, description="The result of the execution"
    )
    has_ended: typing.Optional[bool] = pydantic.Field(
        None, description="If the execution has ended"
    )


class ExecutionMetadata(RuleMetadata):
    node_id: typing.Optional[str] = pydantic.Field(None, description="The node id")
    execution: typing.Optional[ExecutionSchema] = pydantic.Field(
        None, description="Metadata of the rule execution"
    )


class ExceptionSchema(pydantic.BaseModel):
    msg: typing.Any = pydantic.Field(..., description="The exception message")
    exception_type: str = pydantic.Field(..., description="The exception type")
    traceback: typing.Optional[typing.Any] = pydantic.Field(
        None, description="The exception traceback", exclude=True
    )


class DetailSchema(pydantic.BaseModel):
    msg: str = pydantic.Field(
        ...,
        description="A human-readable explanation specific to this occurrence of the problem. This field’s value can be localized",
    )
    type_: str = pydantic.Field(
        ..., alias="type", description="The type of the problem"
    )
    exception: ExceptionSchema = pydantic.Field(
        ..., description="The exception that caused the problem"
    )


class ErrorSchema(pydantic.BaseModel):
    title: str = pydantic.Field(
        ...,
        description="A short, human-readable summary of the problem that SHOULD NOT change from occurrence to occurrence of the problem",
    )
    detail: DetailSchema = pydantic.Field(
        ...,
        description="A explanation specific to this occurrence of the problem. This field’s value can be localized",
    )
    status: str = pydantic.Field(
        ...,
        description="The HTTP status code applicable to this problem, expressed as a string value",
    )
    metadata: ExecutionMetadata = pydantic.Field(..., description="Rule metadata")



import typing

import pandas as pd
import pydantic

from retrack.nodes import BaseNode
from retrack.engine.base import Execution
from retrack.engine.schemas import RuleMetadata
from retrack.engine.request_manager import RequestManager
from retrack.nodes.base import NodeKind, NodeMemoryType
from retrack.utils import constants, exceptions, registry
from retrack.utils.component_registry import ComponentRegistry


class RuleExecutor:
    def __init__(
        self,
        components_registry: ComponentRegistry,
        execution_order: typing.List[str],
        metadata: RuleMetadata,
    ):
        """Class that executes a rule.

        Args:
            components_registry (ComponentRegistry): Components registry.
            execution_order (typing.List[str]): Execution order.
            metadata (RuleMetadata): Rule metadata.

        Raises:
            exceptions.ExecutionException: If there is an error during execution.
            exceptions.ValidationException: If there is an error during validation.
        """
        self._components_registry = components_registry
        self._execution_order = execution_order
        self._metadata = metadata

        input_nodes = self.components_registry.get_by_kind(NodeKind.INPUT)
        input_nodes.extend(self.components_registry.get_by_kind(NodeKind.CONNECTOR))

        self.reset_request_manager(input_nodes)

    @property
    def components_registry(self) -> ComponentRegistry:
        return self._components_registry

    @property
    def execution_order(self) -> typing.List[str]:
        return self._execution_order

    @property
    def metadata(self) -> RuleMetadata:
        return self._metadata

    @property
    def request_manager(self) -> RequestManager:
        return self._request_manager

    @property
    def request_model(self) -> pydantic.BaseModel:
        return self._request_manager.model

    @property
    def constants(self) -> dict:
        return self._constants

    def _set_constants(self):
        constant_nodes = self.components_registry.get_by_memory_type(
            NodeMemoryType.CONSTANT
        )
        self._constants = {}
        for node in constant_nodes:
            for output_connector_name, _ in node.outputs:
                self._constants[f"{node.id}@{output_connector_name}"] = node.data.value

    @property
    def input_columns(self) -> dict:
        return self._input_columns

    def __set_output_connection_filters(
        self,
        node_id: str,
        filter_value: typing.Any,
        execution: Execution,
        connector_filter=None,
    ):
        """If there is a filter, we need to set the children nodes to receive filtered data"""
        if filter_value is None:
            return

        output_connections = self.components_registry.get_node_output_connections(
            node_id, connector_filter=connector_filter
        )
        execution.update_filters(filter_value, output_connections=output_connections)

    def __get_input_params(
        self, node_dict: dict, current_node_filter: pd.Series, execution: Execution
    ) -> dict:
        input_params = {}

        for connector_name, connections in node_dict.get("inputs", {}).items():
            if connector_name.endswith(constants.NULL_SUFFIX):
                continue

            for connection in connections["connections"]:
                input_params[connector_name] = execution.get_state_data(
                    f"{connection['node']}@{connection['output']}",
                    constants=self.constants,
                    filter_by=current_node_filter,
                )

        return input_params

    def __run_node(self, node_id: str, execution: Execution):
        current_node_filter = execution.filters.get(node_id, None)

        self.__set_output_connection_filters(
            node_id, current_node_filter, execution=execution
        )

        node = self.components_registry.get(node_id)

        if node.memory_type == NodeMemoryType.CONSTANT:
            return

        input_params = self.__get_input_params(
            node.model_dump(by_alias=True), current_node_filter, execution=execution
        )

        # TODO: Remove this condition after adding support for kwargs in the run method for all nodes
        if node.kind() == NodeKind.CONNECTOR:
            input_params["context"] = execution.context

        output = node.run(**input_params)

        for output_name, output_value in output.items():
            if (
                output_name == constants.OUTPUT_REFERENCE_COLUMN
                or output_name == constants.OUTPUT_MESSAGE_REFERENCE_COLUMN
            ):  # Setting output values
                execution.set_state_data(output_name, output_value, current_node_filter)
            elif output_name.endswith(constants.FILTER_SUFFIX):  # Setting filters
                self.__set_output_connection_filters(
                    node_id,
                    output_value,
                    execution=execution,
                    connector_filter=output_name,
                )
            else:  # Setting node outputs to be used as inputs by other nodes
                execution.set_state_data(
                    f"{node_id}@{output_name}",
                    output_value,
                    filter_by=current_node_filter,
                )

    def reset_request_manager(self, input_nodes: typing.List[BaseNode]) -> None:
        """Resets the request manager. This method should be called when the input nodes change.

        Args:
            input_nodes (typing.List[BaseNode]): The input nodes to be used.
        """
        self._input_columns = {
            f"{node.id}@{constants.INPUT_OUTPUT_VALUE_CONNECTOR_NAME}": node.data.name
            for node in input_nodes
        }
        self._request_manager = RequestManager(input_nodes)

        self._set_constants()

    def validate_payload(self, payload_df: pd.DataFrame) -> pd.DataFrame:
        """Validates the payload.

        Args:
            payload_df (pd.DataFrame): The payload to be validated.

        Raises:
            exceptions.ValidationException: If there is an error during validation.

        Returns:
            pd.DataFrame: The validated payload.
        """
        if not isinstance(payload_df, pd.DataFrame):
            raise exceptions.ValidationException(
                rule_metadata=self.metadata,
                payload_df=payload_df,
                raised_exception=TypeError("Payload must be a DataFrame"),
                msg="Payload must be a DataFrame",
            )

        try:
            validated = self.request_manager.validate(payload_df.reset_index(drop=True))
        except Exception as e:
            raise exceptions.ValidationException(
                rule_metadata=self.metadata, payload_df=payload_df, raised_exception=e
            )

        return validated

    def execute(
        self,
        payload_df: pd.DataFrame,
        debug_mode: bool = False,
        raise_raw_exception: bool = False,
        context: typing.Optional[registry.Registry] = None,
    ) -> typing.Union[
        pd.DataFrame, typing.Tuple[Execution, typing.Optional[Exception]]
    ]:
        """Executes the rule.

        Args:
            payload_df (pd.DataFrame): The payload to be executed.
            debug_mode (bool, optional): If True, runs the rule in debug mode and returns the exception, if any. Defaults to False.
            raise_raw_exception (bool, optional): If True, raises the raw exception. Defaults to False.
            context (registry.Registry, optional): Global constants to be used during execution. Defaults to None.

        Raises:
            exceptions.ExecutionException: If there is an error during execution.
            exceptions.ValidationException: If there is an error during validation.

        Returns:
            typing.Union[pd.DataFrame, typing.Tuple[Execution, typing.Optional[Exception]]]: The result of the execution or a tuple with the execution and the exception, if any.
        """
        try:
            validated_payload = self.validate_payload(payload_df)
        except exceptions.ValidationException as e:
            if debug_mode:
                return None, e

            raise e

        execution = Execution.from_payload(
            validated_payload=validated_payload,
            input_columns=self.input_columns,
            context=context,
        )

        for node_id in self.execution_order:
            try:
                self.__run_node(node_id, execution=execution)
            except Exception as e:
                if raise_raw_exception:
                    raise e

                msg = None
                if isinstance(e, exceptions.ExecutionException):
                    msg = "Error executing a sub-rule node {} from rule {} version {}".format(
                        node_id, self.metadata.name, self.metadata.version
                    )

                exception = exceptions.ExecutionException(
                    rule_metadata=self.metadata,
                    execution_data=execution.to_model(),
                    node_id=node_id,
                    raised_exception=e,
                    msg=msg,
                )
                if debug_mode:
                    return execution, exception

                raise exception

            if execution.has_ended():
                break

        if debug_mode:
            return execution, None

        return execution.result



import pandas as pd
import pydantic

from retrack.nodes.base import BaseNode, InputConnectionModel, OutputConnectionModel

################################################
# StartsWith Inputs Outputs
################################################


class StartsWithInputsModel(pydantic.BaseModel):
    input_value_0: InputConnectionModel
    input_value_1: InputConnectionModel


class StartsWithOutputsModel(pydantic.BaseModel):
    output_bool: OutputConnectionModel


################################################
# StartsWith Nodes
################################################


class StartsWith(BaseNode):
    inputs: StartsWithInputsModel
    outputs: StartsWithOutputsModel

    def run(self, input_value_0: pd.Series, input_value_1: pd.Series) -> pd.Series:
        return {
            "output_bool": input_value_0.str.startswith(
                input_value_1.to_string(index=False)
            )
        }



import typing

from retrack.nodes.base import BaseNode
from retrack.utils.registry import Registry


class ComponentRegistry(Registry):
    """A registry to store instances of BaseNode (aka Components).

    It also provides indexes to access the value by name, class, kind and memory type."""

    def __init__(self, case_sensitive: bool = False):
        super().__init__(case_sensitive=case_sensitive)
        self._keys_by_name_map = {}
        self._keys_by_class_map = {}
        self._keys_by_kind_map = {}
        self._keys_by_memory_type_map = {}

    ##################
    ### Properties ###
    ##################

    @property
    def keys_by_name_map(self) -> typing.Dict[str, typing.List[str]]:
        return self._keys_by_name_map

    @property
    def keys_by_class_map(self) -> typing.Dict[str, typing.List[str]]:
        return self._keys_by_class_map

    @property
    def keys_by_kind_map(self) -> typing.Dict[str, typing.List[str]]:
        return self._keys_by_kind_map

    @property
    def keys_by_memory_type_map(self) -> typing.Dict[str, typing.List[str]]:
        return self._keys_by_memory_type_map

    ####################
    ### Registration ###
    ####################

    def __register_in_keys_by_name_map(self, key: str, value: BaseNode) -> None:
        name = value.name.lower()
        if name not in self._keys_by_name_map:
            self._keys_by_name_map[name] = []

        self._keys_by_name_map[name].append(key)

    def __register_in_keys_by_class_map(self, key: str, value: BaseNode) -> None:
        class_name = value.__class__.__name__.lower()
        if class_name not in self._keys_by_class_map:
            self._keys_by_class_map[class_name] = []

        self._keys_by_class_map[class_name].append(key)

    def __register_in_keys_by_kind_map(self, key: str, value: BaseNode) -> None:
        node_kind = value.kind()
        if node_kind not in self._keys_by_kind_map:
            self._keys_by_kind_map[node_kind] = []

        self._keys_by_kind_map[node_kind].append(key)

    def __register_in_keys_by_memory_type_map(self, key: str, value: BaseNode) -> None:
        memory_type = value.memory_type()

        if memory_type not in self._keys_by_memory_type_map:
            self._keys_by_memory_type_map[memory_type] = []

        self._keys_by_memory_type_map[memory_type].append(key)

    ######################
    ### Unregistration ###
    ######################

    def __unregister_from_keys_by_name_map(self, key: str, value: BaseNode) -> None:
        name = value.name.lower()
        self._keys_by_name_map[name].remove(key)

    def __unregister_from_keys_by_class_map(self, key: str, value: BaseNode) -> None:
        class_name = value.__class__.__name__.lower()
        self._keys_by_class_map[class_name].remove(key)

    def __unregister_from_keys_by_kind_map(self, key: str, value: BaseNode) -> None:
        node_kind = value.kind()
        self._keys_by_kind_map[node_kind].remove(key)

    def __unregister_from_keys_by_memory_type_map(
        self, key: str, value: BaseNode
    ) -> None:
        memory_type = value.memory_type()
        self._keys_by_memory_type_map[memory_type].remove(key)

    ####################
    ### Public API #####
    ####################

    def register(self, key: str, value: BaseNode, overwrite: bool = False) -> None:
        """Register an entry."""
        if not isinstance(value, BaseNode):
            raise ValueError("value must be a BaseNode instance.")

        super().register(key, value, overwrite=overwrite)

        self.__register_in_keys_by_name_map(key, value)
        self.__register_in_keys_by_class_map(key, value)
        self.__register_in_keys_by_kind_map(key, value)
        self.__register_in_keys_by_memory_type_map(key, value)

    def unregister(self, key: str) -> None:
        """Unregister an entry."""
        if not self._case_sensitive:
            key = key.lower()

        value = self._value.pop(key, None)

        if value is None:
            return

        self.__unregister_from_keys_by_name_map(key, value)
        self.__unregister_from_keys_by_class_map(key, value)
        self.__unregister_from_keys_by_kind_map(key, value)
        self.__unregister_from_keys_by_memory_type_map(key, value)

    #####################
    ### Query methods ###
    #####################

    def get_by_name(self, name: str) -> typing.List[BaseNode]:
        name = name.lower()
        return [self.get(id_) for id_ in self.keys_by_name_map.get(name, [])]

    def get_by_class(self, class_name: str) -> typing.List[BaseNode]:
        class_name = class_name.lower()
        return [self.get(id_) for id_ in self.keys_by_class_map.get(class_name, [])]

    def get_by_kind(self, kind: str) -> typing.List[BaseNode]:
        return [self.get(id_) for id_ in self.keys_by_kind_map.get(kind, [])]

    def get_by_memory_type(self, memory_type: str) -> typing.List[BaseNode]:
        return [
            self.get(id_) for id_ in self.keys_by_memory_type_map.get(memory_type, [])
        ]

    def _filter_connectors(self, connectors, connector_filter) -> typing.List[str]:
        result = []

        for connector_name, value in connectors.items():
            if connector_filter is not None and connector_name != connector_filter:
                continue

            for connection in value["connections"]:
                result.append(connection["node"])
        return result

    def get_node_input_connections(
        self, node_id: str, connector_filter=None
    ) -> typing.List[str]:
        node_dict = self.get(node_id).model_dump(by_alias=True)

        connectors = node_dict.get("inputs", {})
        return self._filter_connectors(connectors, connector_filter)

    def get_node_output_connections(
        self, node_id: str, connector_filter=None
    ) -> typing.List[str]:
        node_dict = self.get(node_id).model_dump(by_alias=True)

        connectors = node_dict.get("outputs", {})
        return self._filter_connectors(connectors, connector_filter)

    def calculate_edges(self) -> typing.List[typing.Tuple[str, str]]:
        edges = []

        for node_id, node in self.memory.items():
            for _, output_connection in node.outputs:
                for c in output_connection.connections:
                    edges.append((node_id, c.node))

        return edges



import typing


class Registry:
    """A simple registry for storing value."""

    def __init__(self, case_sensitive: bool = False):
        self._memory = {}
        self._case_sensitive = case_sensitive

    def register(self, key: str, value: typing.Any, overwrite: bool = False):
        """Register an entry."""
        if not self._case_sensitive:
            key = key.lower()

        if key in self._memory and not overwrite:
            raise ValueError(f"{key} is already registered.")

        self._memory[key] = value

    def unregister(self, key: str):
        """Unregister an entry."""
        if not self._case_sensitive:
            key = key.lower()

        self._memory.pop(key, None)

    def get(self, key: str, default: typing.Any = None) -> typing.Any:
        """Get a registered entry."""
        if not self._case_sensitive:
            key = key.lower()

        return self._memory.get(key, default)

    @property
    def keys(self) -> typing.List[str]:
        """Return the registry keys."""
        return list(self._memory.keys())

    @property
    def values(self) -> typing.List[typing.Any]:
        """Return the registry values."""
        return list(self._memory.values())

    @property
    def memory(self) -> typing.Dict[str, typing.Any]:
        """Return the registry memory."""
        return self._memory

    def __itter__(self):
        return iter(self._memory)

    def __contains__(self, key: str) -> bool:
        """Check if an entry is registered."""
        if not self._case_sensitive:
            key = key.lower()

        return key in self._memory



import json
import typing

import pandas as pd
import pydantic

from retrack.nodes.base import InputConnectionModel, NodeKind, OutputConnectionModel
from retrack.nodes.dynamic.base import BaseDynamicIOModel, BaseDynamicNode
from retrack.utils.registry import Registry


class FlowV0MetadataModel(pydantic.BaseModel):
    value: str
    name: typing.Optional[str] = None
    default: typing.Optional[str] = None

    def parsed_value(self) -> typing.Dict[str, typing.Any]:
        return json.loads(self.value)


class FlowV0OutputsModel(pydantic.BaseModel):
    output_value: OutputConnectionModel


def flow_factory(
    inputs: typing.Dict[str, typing.Any],
    nodes_registry: Registry,
    dynamic_nodes_registry: Registry,
    validator_registry: Registry,
    data: dict,
    rule_class,
    **factory_kwargs,
) -> typing.Type[BaseDynamicNode]:
    graph_data = json.loads(data["value"])
    rule_instance = rule_class.create(
        graph_data=graph_data,
        nodes_registry=nodes_registry,
        dynamic_nodes_registry=dynamic_nodes_registry,
        validator_registry=validator_registry,
        name=data.get("name"),
    )
    input_fields = {}

    connector_nodes = rule_instance.components_registry.get_by_kind(NodeKind.CONNECTOR)

    if len(connector_nodes) > 0:
        raise ValueError(
            f"Flow nodes cannot have connector nodes. Found {len(connector_nodes)} connector nodes."
        )

    for name in inputs.keys():
        input_fields[name] = BaseDynamicNode.create_sub_field(InputConnectionModel)

    inputs_model = BaseDynamicIOModel.with_fields("FlowV0InputsModel", **input_fields)

    models = {
        "inputs": BaseDynamicNode.create_sub_field(inputs_model),
        "outputs": BaseDynamicNode.create_sub_field(FlowV0OutputsModel),
        "data": BaseDynamicNode.create_sub_field(FlowV0MetadataModel),
    }

    BaseFlowV0Model = BaseDynamicNode.with_fields("FlowV0", **models)

    class FlowV0(BaseFlowV0Model):
        def run(self, **kwargs) -> typing.Dict[str, typing.Any]:
            input_args = {}
            for name, value in kwargs.items():
                if name.startswith("input_"):
                    name = name[len("input_") :]

                input_args[name] = value

            response = rule_instance.executor.execute(pd.DataFrame(input_args))

            return {"output_value": response["output"].values}

        def generate_input_nodes(self):
            input_nodes = []
            for component in rule_instance.components_registry.memory.values():
                input_nodes.extend(component.generate_input_nodes())
            return input_nodes

        def kind(self) -> NodeKind:
            return NodeKind.FLOW

    return FlowV0



import json
import typing

from retrack import nodes
from retrack.engine.rule import Rule, RuleExecutor
from retrack.utils import registry


def from_json(
    graph_data: typing.Union[str, dict],
    name: str = None,
    nodes_registry: registry.Registry = nodes.registry(),
    dynamic_nodes_registry: registry.Registry = nodes.dynamic_nodes_registry(),
    return_executor: bool = True,
    **kwargs,
) -> typing.Union[Rule, RuleExecutor]:
    """Create a rule from a json file or a dict.

    Args:
        graph_data (typing.Union[str, dict]): Graph data.
        name (str, optional): Rule name. Defaults to None.
        nodes_registry (registry.Registry, optional): Nodes registry. Defaults to nodes.registry().
        dynamic_nodes_registry (registry.Registry, optional): Dynamic nodes registry. Defaults to nodes.dynamic_nodes_registry().
        return_executor (bool, optional): Whether to return the executor or the rule. Defaults to True.

    Raises:
        ValueError: If the data is not a dict or a json file path.

    Returns:
        typing.Union[Rule, RuleExecutor]: Rule or RuleExecutor depending on return_executor.
    """
    if isinstance(graph_data, str) and graph_data.endswith(".json"):
        if name is None:
            name = graph_data
        graph_data = json.loads(open(graph_data).read())
    elif not isinstance(graph_data, dict):
        raise ValueError("data must be a dict or a json file path")

    rule = Rule.create(
        graph_data=graph_data,
        name=name,
        nodes_registry=nodes_registry,
        dynamic_nodes_registry=dynamic_nodes_registry,
        **kwargs,
    )
    return rule.executor if return_executor else rule



import enum
import typing

import pandas as pd
import pydantic

from retrack.nodes.base import BaseNode, InputConnectionModel, OutputConnectionModel

###############################################################
# Check Metadata Models
###############################################################


class CheckOperator(str, enum.Enum):
    EQUAL = "=="
    NOT_EQUAL = "!="
    GREATER_THAN = ">"
    LESS_THAN = "<"
    GREATER_THAN_OR_EQUAL = ">="
    LESS_THAN_OR_EQUAL = "<="

    def __str__(self):
        return self.value

    def __repr__(self) -> str:
        return self.value


class CheckMetadataModel(pydantic.BaseModel):
    operator: typing.Optional[CheckOperator] = CheckOperator.EQUAL


###############################################################
# Check Inputs and Outputs
###############################################################


class CheckInputsModel(pydantic.BaseModel):
    input_value_0: InputConnectionModel
    input_value_1: InputConnectionModel


class CheckOutputsModel(pydantic.BaseModel):
    output_bool: OutputConnectionModel


###############################################################
# Check Node
###############################################################


class Check(BaseNode):
    data: CheckMetadataModel
    inputs: CheckInputsModel
    outputs: CheckOutputsModel

    def run(
        self,
        input_value_0: pd.Series,
        input_value_1: pd.Series,
    ) -> typing.Dict[str, pd.Series]:
        if self.data.operator == CheckOperator.EQUAL:
            return {
                "output_bool": input_value_0.astype(str) == input_value_1.astype(str)
            }
        elif self.data.operator == CheckOperator.NOT_EQUAL:
            return {
                "output_bool": input_value_0.astype(str) != input_value_1.astype(str)
            }
        elif self.data.operator == CheckOperator.GREATER_THAN:
            return {
                "output_bool": input_value_0.astype(float) > input_value_1.astype(float)
            }
        elif self.data.operator == CheckOperator.LESS_THAN:
            return {
                "output_bool": input_value_0.astype(float) < input_value_1.astype(float)
            }
        elif self.data.operator == CheckOperator.GREATER_THAN_OR_EQUAL:
            return {
                "output_bool": input_value_0.astype(float)
                >= input_value_1.astype(float)
            }
        elif self.data.operator == CheckOperator.LESS_THAN_OR_EQUAL:
            return {
                "output_bool": input_value_0.astype(float)
                <= input_value_1.astype(float)
            }
        else:
            raise ValueError("Unknown operator")



from retrack.utils.registry import Registry
from retrack.validators.base import BaseValidator
from retrack.validators.check_is_dag import CheckIsDAG
from retrack.validators.node_exists import NodeExistsValidator


def registry() -> Registry:
    _registry = Registry()
    _registry.register(
        "single_start_node_exists",
        NodeExistsValidator("start", min_quantity=1, max_quantity=1),
    )
    _registry.register("check_is_dag", CheckIsDAG())
    return _registry


__all__ = ["registry", "BaseValidator"]



from retrack.validators.base import BaseValidator


class NodeExistsValidator(BaseValidator):
    """Validator that checks if a node exists in the graph."""

    def __init__(
        self, node_name: str, max_quantity: int = None, min_quantity: int = None
    ) -> None:
        """Initialize the validator.

        Args:
            node_name: The name of the node to validate.
            max_quantity: The maximum quantity of nodes to validate.
            min_quantity: The minimum quantity of nodes to validate.
        """
        self.node_name = node_name.lower()
        self.max_quantity = max_quantity
        self.min_quantity = min_quantity

    def validate(self, graph_data: dict, **kwargs) -> bool:
        """Validate the graph data.

        Args:
            graph_data: The graph data to validate.

        Returns:
            True if the graph data is valid, False otherwise.
        """
        nodes = graph_data.get("nodes", [])
        nodes = [
            node for _, node in nodes.items() if node["name"].lower() == self.node_name
        ]
        if self.max_quantity is not None and len(nodes) > self.max_quantity:
            return False
        if self.min_quantity is not None and len(nodes) < self.min_quantity:
            return False
        return True



import hashlib
import json

import unicodedata

from retrack.utils.component_registry import ComponentRegistry
from retrack.utils.registry import Registry
from retrack.utils import exceptions
from retrack.engine.schemas import RuleMetadata


def normalize_string(some_string: str) -> str:
    return (
        unicodedata.normalize("NFKD", some_string)
        .encode("ascii", "ignore")
        .decode("utf-8")
    )


def validate_version(
    graph_data: dict, raise_if_null_version: bool, validate_version: bool, name: str
) -> str:
    version = graph_data.get("version", None)

    graph_json_content = (
        json.dumps(graph_data["nodes"], ensure_ascii=False, separators=(",", ":"))
        .replace("\\", "")
        .replace('"', "")
    )
    graph_json_content = normalize_string(graph_json_content)
    calculated_hash = hashlib.sha256(graph_json_content.encode()).hexdigest()[:10]

    if version is None:
        if raise_if_null_version:
            raise exceptions.InvalidVersionException(
                rule_metadata=RuleMetadata(name=name, version=None),
                raised_exception=ValueError("Version is null"),
                msg="Version is null",
            )

        return f"{calculated_hash}.dynamic"

    file_version_hash = version.split(".")[0]

    if file_version_hash != calculated_hash and validate_version:
        raise exceptions.InvalidVersionException(
            rule_metadata=RuleMetadata(name=name, version=version),
            raised_exception=ValueError(
                f"Version hash {file_version_hash} does not match calculated hash {calculated_hash}"
            ),
            msg="Version hash does not match calculated hash",
        )

    return version


def validate_data(data: dict) -> dict:
    if not isinstance(data, dict):
        raise TypeError("Data must be a dictionary. Instead got: " + str(type(data)))
    if "nodes" not in data:
        raise ValueError("No nodes found in data")
    if not isinstance(data["nodes"], dict):
        raise TypeError(
            "Nodes must be a dictionary. Instead got: " + str(type(data["nodes"]))
        )
    return data


def validate_with_validators(
    graph_data: dict, edges: dict, validator_registry: Registry
):
    for validator_name, validator in validator_registry.memory.items():
        if not validator.validate(graph_data=graph_data, edges=edges):
            raise ValueError(f"Invalid graph data: {validator_name}")


def check_node_name(node_name: str, node_id: str):
    if node_name is None:
        raise ValueError(f"Node {node_id} has no name")
    if not isinstance(node_name, str):
        raise TypeError(f"Node {node_id} name must be a string")


def walk(actual_id: str, skiped_ids: list, components_registry: ComponentRegistry):
    skiped_ids.append(actual_id)

    output_ids = components_registry.get_node_output_connections(actual_id)

    for next_id in output_ids:
        if next_id not in skiped_ids:
            next_node_input_ids = components_registry.get_node_input_connections(
                next_id
            )
            run_next = True
            for next_node_input_id in next_node_input_ids:
                if next_node_input_id not in skiped_ids:
                    run_next = False
                    break

            if run_next:
                walk(next_id, skiped_ids, components_registry)

    return skiped_ids


def get_execution_order(components_registry: ComponentRegistry):
    start_nodes = components_registry.get_by_class("start")

    return walk(start_nodes[0].id, [], components_registry)



import pandas as pd
import pydantic

from retrack.nodes.base import BaseNode, InputConnectionModel, OutputConnectionModel
from retrack.utils import transformers

################################################
# EndsWithAny Inputs Outputs
################################################


class EndsWithAnyInputsModel(pydantic.BaseModel):
    input_value: InputConnectionModel
    input_list: InputConnectionModel


class EndsWithAnyOutputsModel(pydantic.BaseModel):
    output_bool: OutputConnectionModel


################################################
# EndsWithAny Nodes
################################################


class EndsWithAny(BaseNode):
    inputs: EndsWithAnyInputsModel
    outputs: EndsWithAnyOutputsModel

    def run(self, input_value: pd.Series, input_list: pd.Series) -> pd.Series:
        input_list = transformers.to_list(input_list)
        return {"output_bool": input_value.str.endswith(tuple(input_list))}



import json
import typing

import pandas as pd
import numpy as np
import pydantic

from retrack.nodes.base import InputConnectionModel, OutputConnectionModel
from retrack.nodes.dynamic.base import BaseDynamicIOModel, BaseDynamicNode


LINK_FUNCS = {
    "identity": lambda eta: eta,
    "log": np.log,
    "exponential": np.exp,
    "logit": lambda eta: np.log(eta / (1 - eta)),
    "inverse": lambda eta: 1.0 / eta,
}


class GLMMetadataModel(pydantic.BaseModel):
    value: str
    name: typing.Optional[str] = None
    default: typing.Optional[str] = None
    link: typing.Optional[str] = "log"
    headers_map: typing.Dict[str, int]

    def parsed_value(self) -> typing.Dict[str, typing.Any]:
        return json.loads(self.value)

    def intercept(self) -> float:
        parsed = self.parsed_value()
        return float(parsed.get("intercept", 0.0))


class GLMOutputsModel(pydantic.BaseModel):
    output_value: OutputConnectionModel


def glm_factory(
    inputs: typing.Dict[str, typing.Any],
    **factory_kwargs,
) -> typing.Type[BaseDynamicNode]:
    input_fields = {}

    for name in inputs.keys():
        input_fields[name] = BaseDynamicNode.create_sub_field(InputConnectionModel)

    inputs_model = BaseDynamicIOModel.with_fields("GLMInputsModel", **input_fields)

    models = {
        "inputs": BaseDynamicNode.create_sub_field(inputs_model),
        "outputs": BaseDynamicNode.create_sub_field(GLMOutputsModel),
        "data": BaseDynamicNode.create_sub_field(GLMMetadataModel),
    }

    BaseGLMModel = BaseDynamicNode.with_fields("GLM", **models)

    class GLM(BaseGLMModel):
        def run(self, **kwargs) -> typing.Dict[str, typing.Any]:
            weights = self.data.parsed_value()

            intercept = self.data.intercept()

            _example_input: pd.Series = kwargs.get("input_value_0")
            dot = pd.Series(
                np.zeros(len(_example_input)), dtype=float, index=_example_input.index
            )

            for feature_name, index in self.data.headers_map.items():
                field_name = f"input_value_{index}"
                if field_name not in kwargs:
                    raise ValueError(f"Missing input {field_name} in GLM node")

                if feature_name not in weights:
                    raise ValueError(
                        f"Missing weight for feature {feature_name} in GLM node"
                    )

                dot += kwargs[field_name].astype(float) * float(weights[feature_name])

            response = LINK_FUNCS[self.data.link](intercept + dot.astype(float))

            return {"output_value": response}

    return GLM



import typing

import pydantic

from retrack.nodes.base import (
    BaseNode,
    InputConnectionModel,
    NodeKind,
    OutputConnectionModel,
    OptionalCastedToNoneStringType,
)

################################################
# Input Metadata Models
################################################


class InputMetadataModel(pydantic.BaseModel):
    name: str
    default: OptionalCastedToNoneStringType = None


################################################
# Input Inputs and Outputs
################################################


class InputInputsModel(pydantic.BaseModel):
    input_void: typing.Optional[InputConnectionModel] = None


class InputOutputsModel(pydantic.BaseModel):
    output_value: OutputConnectionModel


################################################
# Input Node
################################################


class Input(BaseNode):
    data: InputMetadataModel
    inputs: typing.Optional[InputInputsModel] = None
    outputs: InputOutputsModel

    def kind(self) -> NodeKind:
        return NodeKind.INPUT



import networkx as nx

from retrack.validators.base import BaseValidator


class CheckIsDAG(BaseValidator):
    def validate(self, edges: list, **kwargs) -> bool:
        """Validate the graph data.

        Returns:
            True if the graph data is valid, False otherwise.
        """
        graph = nx.DiGraph()
        graph.add_edges_from(edges)
        return nx.is_directed_acyclic_graph(graph)



import typing

import pandas as pd
import pydantic

from retrack.nodes.base import InputConnectionModel, OutputConnectionModel
from retrack.nodes.dynamic.base import BaseDynamicIOModel, BaseDynamicNode


class CSVTableV0MetadataModel(pydantic.BaseModel):
    value: typing.Union[typing.List[str], typing.List[typing.List[str]]]
    target: str
    headers: typing.List[str]
    headers_map: typing.List[str]
    separator: typing.Optional[str] = ","
    default: typing.Optional[str] = None

    def df(self) -> pd.DataFrame:
        rows = [values.split(self.separator) for values in self.value[1:]]
        return pd.DataFrame(rows, columns=self.headers_map)


class CSVTableV0OutputsModel(pydantic.BaseModel):
    output_value: OutputConnectionModel


def csv_table_factory(
    inputs: typing.Dict[str, typing.Any], **kwargs
) -> typing.Type[BaseDynamicNode]:
    input_fields = {}

    for name in inputs.keys():
        input_fields[name] = BaseDynamicNode.create_sub_field(InputConnectionModel)

    inputs_model = BaseDynamicIOModel.with_fields(
        "CSVTableV0InputsModel", **input_fields
    )

    models = {
        "inputs": BaseDynamicNode.create_sub_field(inputs_model),
        "outputs": BaseDynamicNode.create_sub_field(CSVTableV0OutputsModel),
        "data": BaseDynamicNode.create_sub_field(CSVTableV0MetadataModel),
    }

    BaseCSVTableV0Model = BaseDynamicNode.with_fields("CSVTableV0", **models)

    class CSVTableV0(BaseCSVTableV0Model):
        def run(self, **kwargs) -> typing.Dict[str, typing.Any]:
            csv_df = self.data.df()

            response_df = {}

            input_columns = [
                name for name in self.data.headers_map if name != self.data.target
            ]

            for name in input_columns:
                if name == self.data.target:
                    continue

                if name not in kwargs.keys():
                    raise ValueError(f"Missing input {name} in CSVTableV0 node")

                response_df[name] = kwargs[name]

            response_df = pd.DataFrame(response_df)
            response_df = response_df.astype(str)
            response_df = response_df.merge(csv_df, how="left", on=input_columns)

            if self.data.default:
                response_df[self.data.target] = response_df[self.data.target].fillna(
                    self.data.default
                )

            response_df.set_index(
                kwargs[input_columns[0]].index, inplace=True, drop=True
            )

            return {"output_value": response_df[self.data.target]}

    return CSVTableV0



class BaseValidator:
    """Base class for all validators."""

    def validate(self, **kwargs) -> bool:
        """Validate the graph data.

        Returns:
            True if the graph data is valid, False otherwise.
        """
        raise NotImplementedError



import typing

import numpy as np
import pandas as pd
import pydantic

from retrack.nodes.base import (
    BaseNode,
    InputConnectionModel,
    NodeKind,
    NodeMemoryType,
    OutputConnectionModel,
)

#######################################################
# Constant Metadata Models
#######################################################


class ConstantMetadataModel(pydantic.BaseModel):
    value: str


class ListMetadataModel(pydantic.BaseModel):
    value: typing.List[str]


class BoolMetadataModel(pydantic.BaseModel):
    value: typing.Optional[bool] = pydantic.Field(False, alias="value")

    @pydantic.field_validator("value")
    def validate_value(cls, value):
        if value is None:
            return False

        value = str(value).lower()

        if value in ["true", "1", "yes", "y", "t"]:
            return True

        return False


class IntervalCatMetadataModel(pydantic.BaseModel):
    value: typing.List[str]
    start_interval_column: str
    end_interval_column: str
    category_column: str
    headers: typing.List[str]
    separator: typing.Optional[str] = pydantic.Field(",")
    default: typing.Optional[str] = None

    def df(self) -> pd.DataFrame:
        rows = [values.split(self.separator) for values in self.value[1:]]
        df = pd.DataFrame(rows, columns=self.headers)

        df[self.start_interval_column] = df[self.start_interval_column].astype(float)
        df[self.end_interval_column] = df[self.end_interval_column].astype(float)

        return df


#######################################################
# Constant Inputs and Outputs
#######################################################


class ConstantInputsModel(pydantic.BaseModel):
    input_void: typing.Optional[InputConnectionModel] = None


class ConstantInputsValueModel(pydantic.BaseModel):
    input_value: InputConnectionModel


class ConstantOutputsModel(pydantic.BaseModel):
    output_value: OutputConnectionModel


class ListOutputsModel(pydantic.BaseModel):
    output_list: OutputConnectionModel


class BoolOutputsModel(pydantic.BaseModel):
    output_bool: OutputConnectionModel


#######################################################
# Constant Nodes
#######################################################


class BaseConstant(BaseNode):
    inputs: typing.Optional[ConstantInputsModel] = None

    def kind(self) -> NodeKind:
        return NodeKind.CONSTANT


class Constant(BaseConstant):
    data: ConstantMetadataModel
    outputs: ConstantOutputsModel

    def run(self, **kwargs) -> typing.Dict[str, typing.Any]:
        return {"output_value": self.data.value}


class List(BaseConstant):
    data: ListMetadataModel
    outputs: ListOutputsModel

    def run(self, **kwargs) -> typing.Dict[str, typing.Any]:
        return {}  # {"output_list": self.data.value}

    def memory_type(self) -> NodeMemoryType:
        return NodeMemoryType.CONSTANT


class Bool(BaseConstant):
    data: BoolMetadataModel = BoolMetadataModel(value=False)
    outputs: BoolOutputsModel

    def run(self, **kwargs) -> typing.Dict[str, typing.Any]:
        return {"output_bool": self.data.value}


class IntervalCatV0(BaseConstant):
    data: IntervalCatMetadataModel
    inputs: ConstantInputsValueModel
    outputs: ConstantOutputsModel

    def run(self, input_value: pd.Series) -> typing.Dict[str, typing.Any]:
        values = input_value.astype(float).copy()
        output = pd.Series(np.nan, index=input_value.index, dtype="object")

        for _, row in self.data.df().iterrows():
            output.loc[
                (values >= float(row[self.data.start_interval_column]))
                & (values < float(row[self.data.end_interval_column])),
            ] = row[self.data.category_column]

        replace_value_tag = output == "{value}"
        output.loc[replace_value_tag] = input_value[replace_value_tag]

        if self.data.default is not None:
            output.fillna(self.data.default, inplace=True)

        return {"output_value": output}



from retrack.nodes.base import NodeKind
from retrack.nodes.inputs import Input, InputMetadataModel


class ConnectorMetadataModel(InputMetadataModel):
    service: str
    identifier: str


class BaseConnector(Input):
    data: ConnectorMetadataModel

    def kind(self) -> NodeKind:
        return NodeKind.CONNECTOR

    def run(self, **kwargs):
        return {}



import typing

import pandas as pd
import pydantic

from retrack.nodes.base import BaseNode, InputConnectionModel, OutputConnectionModel


class LowerCaseOutputsModel(pydantic.BaseModel):
    output_value: OutputConnectionModel


class LowerCaseInputsModel(pydantic.BaseModel):
    input_value: InputConnectionModel


class LowerCase(BaseNode):
    inputs: LowerCaseInputsModel
    outputs: LowerCaseOutputsModel

    def run(
        self,
        input_value: pd.Series,
    ) -> typing.Dict[str, pd.Series]:
        return {"output_value": input_value.astype(str).str.lower()}



import typing

import pandas as pd
import pydantic

from retrack.nodes.base import (
    BaseNode,
    InputConnectionModel,
    NodeKind,
    OptionalCastedToNoneStringType,
)
from retrack.utils import constants

################################################
# Output Metadata Models
################################################


class OutputMetadataModel(pydantic.BaseModel):
    message: OptionalCastedToNoneStringType = None


################################################
# Output Inputs and Outputs
################################################
class OutputInputsModel(pydantic.BaseModel):
    input_value: InputConnectionModel


################################################
# Output Node
################################################


class Output(BaseNode):
    inputs: typing.Optional[OutputInputsModel]
    data: OutputMetadataModel

    def kind(self) -> NodeKind:
        return NodeKind.OUTPUT

    def run(self, input_value: pd.Series) -> typing.Dict[str, pd.Series]:
        return {
            constants.OUTPUT_REFERENCE_COLUMN: input_value,
            constants.OUTPUT_MESSAGE_REFERENCE_COLUMN: self.data.message,
        }



from retrack.engine.rule import Rule
from retrack.engine.constructor import from_json
from retrack.engine.executor import RuleExecutor
from retrack.engine.base import Execution
from retrack.nodes import registry as nodes_registry, dynamic_nodes_registry
from retrack.nodes.base import BaseNode, InputConnectionModel, OutputConnectionModel
from retrack.nodes.connectors import BaseConnector

__all__ = [
    "Rule",
    "from_json",
    "RuleExecutor",
    "Execution",
    "nodes_registry",
    "dynamic_nodes_registry",
    "BaseNode",
    "InputConnectionModel",
    "OutputConnectionModel",
    "BaseConnector",
]



import datetime as dt
import typing

import pandas as pd
import pydantic

from retrack.nodes.base import BaseNode, InputConnectionModel, OutputConnectionModel

###############################################################
# CurrentYear Inputs and Outputs
###############################################################


class CurrentYearInputsModel(pydantic.BaseModel):
    input_void: typing.Optional[InputConnectionModel] = None


class CurrentYearOutputsModel(pydantic.BaseModel):
    output_value: OutputConnectionModel


###############################################################
# CurrentYear Node
###############################################################


class CurrentYear(BaseNode):
    inputs: CurrentYearInputsModel
    outputs: CurrentYearOutputsModel

    def run(
        self,
        input_void: typing.Optional[pd.Series] = None,
    ) -> typing.Dict[str, str]:
        return {"output_value": dt.datetime.now().year}



import typing

import numpy as np
import pandas as pd

from retrack.utils import constants, registry
from retrack.engine.schemas import ExecutionSchema


class Execution:
    def __init__(
        self,
        payload: pd.DataFrame,
        states: pd.DataFrame,
        filters: dict = None,
        context: registry.Registry = None,
    ):
        self.payload = payload
        self.states = states
        self.filters = filters or {}
        self.context = context

    def set_state_data(
        self, column: str, value: typing.Any, filter_by: typing.Any = None
    ):
        if filter_by is None:
            self.states[column] = value
        else:
            self.states.loc[filter_by, column] = value

    def get_state_data(
        self, column: str, constants: dict, filter_by: typing.Any = None
    ):
        if column in constants:
            return constants[column]

        if filter_by is None:
            return self.states[column]

        return self.states.loc[filter_by, column]

    def update_filters(self, filter_value, output_connections: typing.List[str] = None):
        for output_connection_id in output_connections:
            if self.filters.get(output_connection_id, None) is None:
                self.filters[output_connection_id] = filter_value
            else:
                self.filters[output_connection_id] = (
                    self.filters[output_connection_id] & filter_value
                )

    @classmethod
    def from_payload(
        cls,
        validated_payload: pd.DataFrame,
        input_columns: dict,
        context: registry.Registry = None,
    ):
        state_df = pd.DataFrame([])
        for node_id, input_name in input_columns.items():
            state_df[node_id] = validated_payload[input_name].copy()

        state_df[constants.OUTPUT_REFERENCE_COLUMN] = np.nan
        state_df[constants.OUTPUT_MESSAGE_REFERENCE_COLUMN] = np.nan

        return cls(
            payload=validated_payload,
            states=state_df,
            context=context,
        )

    @property
    def result(self) -> pd.DataFrame:
        return self.states[
            [
                constants.OUTPUT_REFERENCE_COLUMN,
                constants.OUTPUT_MESSAGE_REFERENCE_COLUMN,
            ]
        ]

    def has_ended(self) -> bool:
        return self.states[constants.OUTPUT_REFERENCE_COLUMN].isna().sum() == 0

    def to_dict(self) -> dict:
        return {
            "payload": self.payload.to_dict(),
            "states": self.states.to_dict(),
            "filters": {k: v.to_dict() for k, v in self.filters.items()},
            "result": self.result.to_dict(),
            "has_ended": self.has_ended(),
        }

    def to_model(self) -> ExecutionSchema:
        return ExecutionSchema(**self.to_dict())

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            payload=pd.DataFrame(data["payload"]),
            states=pd.DataFrame(data["states"]),
            filters={k: pd.DataFrame(v) for k, v in data["filters"].items()},
        )

    def __repr__(self) -> str:
        return f"Execution({self.to_dict()})"

    def __str__(self) -> str:
        return self.__repr__()



import pandas as pd
import pydantic

from retrack.nodes.base import BaseNode, InputConnectionModel, OutputConnectionModel

################################################
# EndsWith Inputs Outputs
################################################


class EndsWithInputsModel(pydantic.BaseModel):
    input_value_0: InputConnectionModel
    input_value_1: InputConnectionModel


class EndsWithOutputsModel(pydantic.BaseModel):
    output_bool: OutputConnectionModel


################################################
# EndsWith Nodes
################################################


class EndsWith(BaseNode):
    inputs: EndsWithInputsModel
    outputs: EndsWithOutputsModel

    def run(self, input_value_0: pd.Series, input_value_1: pd.Series) -> pd.Series:
        return {
            "output_bool": input_value_0.str.endswith(
                input_value_1.to_string(index=False)
            )
        }



import typing

import pydantic

from retrack import validators
from retrack.engine.schemas import RuleMetadata
from retrack.engine.executor import RuleExecutor
from retrack.utils import graph
from retrack.utils.component_registry import ComponentRegistry
from retrack.utils.registry import Registry


class Rule(RuleMetadata):
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    components_registry: ComponentRegistry
    execution_order: typing.List[str]
    _executor: RuleExecutor = None

    @property
    def executor(self) -> RuleExecutor:
        if self._executor is None:
            self._executor = RuleExecutor(
                self.components_registry,
                self.execution_order,
                self.as_metadata(),
            )
        return self._executor

    def as_metadata(self) -> RuleMetadata:
        return RuleMetadata(**self.model_dump())

    @classmethod
    def create(
        cls,
        graph_data: dict,
        nodes_registry: Registry,
        dynamic_nodes_registry: Registry,
        validator_registry: Registry = validators.registry(),
        raise_if_null_version: bool = False,
        validate_version: bool = False,
        name: str = None,
    ):
        components_registry = Rule.create_component_registry(
            graph_data, nodes_registry, dynamic_nodes_registry, validator_registry
        )
        version = graph.validate_version(
            graph_data, raise_if_null_version, validate_version, name
        )
        graph_data = graph_data

        graph.validate_with_validators(
            graph_data,
            components_registry.calculate_edges(),
            validator_registry,
        )

        execution_order = graph.get_execution_order(components_registry)

        return cls(
            version=version,
            components_registry=components_registry,
            execution_order=execution_order,
            name=name,
        )

    @staticmethod
    def create_component_registry(
        graph_data: dict,
        nodes_registry: Registry,
        dynamic_nodes_registry: Registry,
        validator_registry: Registry,
    ) -> ComponentRegistry:
        components_registry = ComponentRegistry()
        graph_data = graph.validate_data(graph_data)
        for node_id, node_metadata in graph_data["nodes"].items():
            if node_id in components_registry:
                raise ValueError(f"Duplicate node id: {node_id}")

            node_name = node_metadata.get("name", None)
            graph.check_node_name(node_name, node_id)

            node_name = node_name.lower()

            node_factory = dynamic_nodes_registry.get(node_name)

            if node_factory is not None:
                validation_model = node_factory(
                    **node_metadata,
                    nodes_registry=nodes_registry,
                    dynamic_nodes_registry=dynamic_nodes_registry,
                    validator_registry=validator_registry,
                    rule_class=Rule,
                )
            else:
                validation_model = nodes_registry.get(node_name)

            if validation_model is None:
                raise ValueError(f"Unknown node name: {node_name}")

            component = validation_model(**node_metadata)

            components_registry.register(node_id, validation_model(**node_metadata))

            for input_node in component.generate_input_nodes():
                components_registry.register(input_node.id, input_node, overwrite=True)

        return components_registry



import typing

import pydantic

from retrack.nodes.base import BaseNode


class BaseDynamicIOModel(pydantic.BaseModel):
    @classmethod
    def with_fields(cls, class_name: str, **field_definitions):
        return pydantic.create_model(
            class_name, __base__=BaseDynamicIOModel, **field_definitions
        )


class BaseDynamicNode(BaseNode):
    @classmethod
    def with_fields(cls, class_name: str, **field_definitions):
        return pydantic.create_model(class_name, __base__=BaseNode, **field_definitions)

    @staticmethod
    def create_sub_field(
        type_: typing.Any, optional: bool = False, default_value: typing.Any = None
    ) -> typing.Tuple[typing.Type, typing.Any]:
        if optional:
            return (typing.Optional[type_], default_value)

        return (type_, ...)



from retrack.nodes.dynamic.base import BaseDynamicNode
from retrack.nodes.dynamic.csv_table import csv_table_factory
from retrack.nodes.dynamic.flow import flow_factory
from retrack.nodes.dynamic.conditional_connector import conditional_connector_factory
from retrack.nodes.dynamic.glm import glm_factory
from retrack.utils.registry import Registry


def registry() -> Registry:
    """Create a registry with all the dynamic nodes available in the library.

    A dynamic node is a node that is not explicitly defined in the nodes registry, but is created dynamically from a factory function."""
    _registry = Registry()

    _registry.register("CSVTableV0", csv_table_factory)
    _registry.register("FlowV0", flow_factory)
    _registry.register("ConditionalConnector", conditional_connector_factory)
    _registry.register("BureauConnector", conditional_connector_factory)
    _registry.register("ModelConnector", conditional_connector_factory)
    _registry.register("FeatureConnector", conditional_connector_factory)
    _registry.register("GLM", glm_factory)

    return _registry


__all__ = ["registry", "BaseDynamicNode"]



import typing

import pandas as pd
import pydantic

from retrack.nodes.base import (
    BaseNode,
    InputConnectionModel,
    NodeKind,
    NodeMemoryType,
    OutputConnectionModel,
)

################################################
# If Inputs and Outputs
################################################


class IfInputsModel(pydantic.BaseModel):
    input_bool: InputConnectionModel


class IfOutputsModel(pydantic.BaseModel):
    output_then_filter: OutputConnectionModel
    output_else_filter: OutputConnectionModel


################################################
# If Node
################################################


class If(BaseNode):
    inputs: IfInputsModel
    outputs: IfOutputsModel

    def kind(self) -> NodeKind:
        return NodeKind.FILTER

    def run(self, input_bool: pd.Series) -> typing.Dict[str, pd.Series]:
        return {
            "output_then_filter": input_bool,
            "output_else_filter": ~input_bool,
        }

    def memory_type(self) -> NodeMemoryType:
        return NodeMemoryType.FILTER



import pandas as pd
import pydantic

from retrack.nodes.base import BaseNode, InputConnectionModel, OutputConnectionModel
from retrack.utils import transformers

################################################
# Contains Inputs Outputs
################################################


class ContainsInputsModel(pydantic.BaseModel):
    input_list: InputConnectionModel
    input_value: InputConnectionModel


class ContainsOutputsModel(pydantic.BaseModel):
    output_bool: OutputConnectionModel


################################################
# Contains Nodes
################################################


class Contains(BaseNode):
    inputs: ContainsInputsModel
    outputs: ContainsOutputsModel

    def run(self, input_list: pd.Series, input_value: pd.Series) -> pd.Series:
        return {"output_bool": input_value.isin(transformers.to_list(input_list))}



OUTPUT_REFERENCE_COLUMN = "output"
OUTPUT_MESSAGE_REFERENCE_COLUMN = "message"
NULL_SUFFIX = "_void"
FILTER_SUFFIX = "_filter"
INPUT_OUTPUT_VALUE_CONNECTOR_NAME = "output_value"
FLOW_NODE_NAME = "flowv0"



import typing

from retrack.nodes.base import NodeKind
import pydantic

from retrack.nodes.base import InputConnectionModel, OutputConnectionModel
from retrack.nodes.dynamic.base import BaseDynamicIOModel, BaseDynamicNode


class ConditionalConnectorMetadataModel(pydantic.BaseModel):
    name: str
    source: str
    resource: str
    default: typing.Optional[str] = None
    headers: typing.List[str]
    headers_map: typing.List[str]


class ConditionalConnectorOutputsModel(pydantic.BaseModel):
    output_value: OutputConnectionModel


def conditional_connector_factory(
    inputs: typing.Dict[str, typing.Any], **kwargs
) -> typing.Type[BaseDynamicNode]:
    input_fields = {}

    for name in inputs.keys():
        input_fields[name] = BaseDynamicNode.create_sub_field(InputConnectionModel)

    inputs_model = BaseDynamicIOModel.with_fields(
        "ConditionalConnectorInputsModel", **input_fields
    )

    models = {
        "inputs": BaseDynamicNode.create_sub_field(inputs_model),
        "outputs": BaseDynamicNode.create_sub_field(ConditionalConnectorOutputsModel),
        "data": BaseDynamicNode.create_sub_field(ConditionalConnectorMetadataModel),
    }

    BaseModel = BaseDynamicNode.with_fields("ConditionalConnectorBaseModel", **models)

    class ConditionalConnector(BaseModel):
        def kind(self) -> NodeKind:
            return NodeKind.CONNECTOR

        def run(self, **kwargs):
            return {}

    return ConditionalConnector



import enum
import typing

import pandas as pd
import pydantic

from retrack.nodes.base import BaseNode, InputConnectionModel, OutputConnectionModel

###############################################################
# Math Metadata Models
###############################################################


class MathOperator(str, enum.Enum):
    SUM = "+"
    SUB = "-"
    DIVISION = "/"
    MULTIPLY = "*"


class MathMetadataModel(pydantic.BaseModel):
    operator: typing.Optional[MathOperator] = MathOperator.SUM


###############################################################
# Math Inputs and Outputs
###############################################################


class MathInputsModel(pydantic.BaseModel):
    input_value_0: InputConnectionModel
    input_value_1: InputConnectionModel


class MathOutputsModel(pydantic.BaseModel):
    output_value: OutputConnectionModel


###############################################################
# Math Node
###############################################################


class Math(BaseNode):
    data: MathMetadataModel
    inputs: MathInputsModel
    outputs: MathOutputsModel

    def run(
        self,
        input_value_0: pd.Series,
        input_value_1: pd.Series,
    ) -> typing.Dict[str, pd.Series]:
        if self.data.operator == MathOperator.SUM:
            return {
                "output_value": input_value_0.astype(float)
                + input_value_1.astype(float)
            }
        elif self.data.operator == MathOperator.SUB:
            return {
                "output_value": input_value_0.astype(float)
                - input_value_1.astype(float)
            }
        elif self.data.operator == MathOperator.MULTIPLY:
            return {
                "output_value": input_value_0.astype(float)
                * input_value_1.astype(float)
            }
        elif self.data.operator == MathOperator.DIVISION:
            return {
                "output_value": input_value_0.astype(float)
                / input_value_1.astype(float)
            }
        else:
            raise ValueError("Unknown operator")


###############################################################
# Absolute Value Node
###############################################################


class AbsoluteValueInputsModel(pydantic.BaseModel):
    input_value: InputConnectionModel


class AbsoluteValue(BaseNode):
    inputs: AbsoluteValueInputsModel
    outputs: MathOutputsModel

    def run(
        self,
        input_value: pd.Series,
    ) -> typing.Dict[str, pd.Series]:
        return {"output_value": input_value.astype(float).abs()}


###############################################################
# Round Node
###############################################################


class Round(BaseNode):
    inputs: AbsoluteValueInputsModel
    outputs: MathOutputsModel

    def run(
        self,
        input_value: pd.Series,
    ) -> typing.Dict[str, pd.Series]:
        return {"output_value": input_value.astype(float).round(0).astype(int)}



import pydantic

from retrack.nodes.base import BaseNode, NodeKind, OutputConnectionModel

################################################
# Start Inputs and Outputs
################################################


class StartOutputsModel(pydantic.BaseModel):
    output_up_void: OutputConnectionModel
    output_down_void: OutputConnectionModel


################################################
# Start Node
################################################


class Start(BaseNode):
    outputs: StartOutputsModel

    def kind(self) -> NodeKind:
        return NodeKind.START



import typing

import pandas as pd
import pydantic

from retrack.nodes.base import BaseNode, InputConnectionModel, OutputConnectionModel


class GetCharOutputsModel(pydantic.BaseModel):
    output_value: OutputConnectionModel


class GetCharInputsModel(pydantic.BaseModel):
    input_value: InputConnectionModel


class GetCharMetadataModel(pydantic.BaseModel):
    index: int


class GetChar(BaseNode):
    inputs: GetCharInputsModel
    outputs: GetCharOutputsModel
    data: GetCharMetadataModel

    def run(
        self,
        input_value: pd.Series,
    ) -> typing.Dict[str, pd.Series]:
        return {
            "output_value": input_value.apply(lambda x: str(x)[self.data.index - 1])
        }



import pandas as pd
import pydantic

from retrack.nodes.base import BaseNode, InputConnectionModel, OutputConnectionModel

################################################
# IsSubStringOf Inputs Outputs
################################################


class IsSubStringOfInputsModel(pydantic.BaseModel):
    input_value_0: InputConnectionModel
    input_value_1: InputConnectionModel


class IsSubStringOfOutputsModel(pydantic.BaseModel):
    output_bool: OutputConnectionModel


################################################
# IsSubStringOf Nodes
################################################


class IsSubStringOf(BaseNode):
    inputs: IsSubStringOfInputsModel
    outputs: IsSubStringOfOutputsModel

    def run(self, input_value_0: pd.Series, input_value_1: pd.Series) -> pd.Series:
        tmp_df = pd.DataFrame(
            {"input_value_0": input_value_0, "input_value_1": input_value_1}
        )

        return {
            "output_bool": tmp_df.apply(
                lambda x: x["input_value_0"] in x["input_value_1"], axis=1
            )
        }



import pandas as pd
import traceback
from retrack.engine.schemas import (
    ErrorSchema,
    DetailSchema,
    ExceptionSchema,
    ExecutionMetadata,
    RuleMetadata,
    ExecutionSchema,
)


def create_exception_schema(raised_exception: Exception) -> ExceptionSchema:
    traceback_str = None
    try:
        traceback_str = traceback.format_exception(raised_exception)
    except Exception:
        pass

    raised_message = str(raised_exception)
    if hasattr(raised_exception, "error"):
        raised_message = raised_exception.error.model_dump()

    return ExceptionSchema(
        msg=raised_message,
        exception_type=type(raised_exception).__name__,
        traceback=traceback_str,
    )


class ExecutionException(Exception):
    """Exception raised when an error occurs during execution of a command."""

    def __init__(
        self,
        rule_metadata: RuleMetadata,
        execution_data: ExecutionSchema,
        node_id: int,
        raised_exception: Exception,
        msg: str = None,
    ):
        msg = (
            msg
            or f"Error executing node {node_id} from rule {rule_metadata.name} version {rule_metadata.version}"
        )

        self.error = ErrorSchema(
            title="Execution error",
            status="500",
            detail=DetailSchema(
                msg=msg,
                type="ExecutionException",
                exception=create_exception_schema(raised_exception),
            ),
            metadata=ExecutionMetadata(
                name=rule_metadata.name,
                version=rule_metadata.version,
                node_id=node_id,
                execution=execution_data,
            ),
        )

        super().__init__(self.error.model_dump())


class ValidationException(Exception):
    """Exception raised when an error occurs during validation of a command."""

    def __init__(
        self,
        rule_metadata: RuleMetadata,
        payload_df: pd.DataFrame,
        raised_exception: Exception,
        msg: str = None,
    ):
        msg = (
            msg
            or f"Error validating payload for rule {rule_metadata.name} version {rule_metadata.version}"
        )

        self.error = ErrorSchema(
            title="Validation error",
            status="400",
            detail=DetailSchema(
                msg=msg,
                type="ValidationException",
                exception=create_exception_schema(raised_exception),
            ),
            metadata=ExecutionMetadata(
                name=rule_metadata.name,
                version=rule_metadata.version,
                node_id=None,
                execution=ExecutionSchema(payload=str(payload_df)),
            ),
        )

        super().__init__(self.error.model_dump())


class InvalidVersionException(Exception):
    """Exception raised when an invalid version is found."""

    def __init__(
        self, rule_metadata: RuleMetadata, raised_exception: Exception, msg: str = None
    ):
        msg = (
            msg
            or f"Invalid version for rule {rule_metadata.name} version {rule_metadata.version}"
        )

        self.error = ErrorSchema(
            title="Invalid version",
            status="400",
            detail=DetailSchema(
                msg=msg,
                type="InvalidVersionException",
                exception=create_exception_schema(raised_exception),
            ),
            metadata=ExecutionMetadata(
                name=rule_metadata.name,
                version=rule_metadata.version,
                node_id=None,
                execution=None,
            ),
        )

        super().__init__(self.error.model_dump())



from retrack.nodes.base import BaseNode
from retrack.nodes.check import Check
from retrack.nodes.connectors import BaseConnector
from retrack.nodes.constants import Bool, Constant, IntervalCatV0, List
from retrack.nodes.contains import Contains
from retrack.nodes.datetime import CurrentYear
from retrack.nodes.dynamic import BaseDynamicNode
from retrack.nodes.dynamic import registry as dynamic_nodes_registry
from retrack.nodes.endswith import EndsWith
from retrack.nodes.endswithany import EndsWithAny
from retrack.nodes.inputs import Input
from retrack.nodes.logic import And, Not, Or
from retrack.nodes.lowercase import LowerCase
from retrack.nodes.match import If
from retrack.nodes.math import AbsoluteValue, Math, Round
from retrack.nodes.outputs import Output
from retrack.nodes.start import Start
from retrack.nodes.startswith import StartsWith
from retrack.nodes.startswithany import StartsWithAny
from retrack.nodes.substring import IsSubStringOf
from retrack.nodes.getchar import GetChar
from retrack.utils.registry import Registry


def registry() -> Registry:
    """Create a registry with all the nodes available in the library."""
    _registry = Registry()

    _registry.register("Input", Input)
    _registry.register("Connector", BaseConnector)  # By default, Connector is an Input
    _registry.register(
        "ConnectorV0", BaseConnector
    )  # By default, Connector is an Input
    _registry.register("Start", Start)
    _registry.register("Constant", Constant)
    _registry.register("List", List)
    _registry.register("Bool", Bool)
    _registry.register("Output", Output)
    _registry.register("Check", Check)
    _registry.register("If", If)
    _registry.register("And", And)
    _registry.register("Or", Or)
    _registry.register("Not", Not)
    _registry.register("Math", Math)
    _registry.register("Round", Round)
    _registry.register("AbsoluteValue", AbsoluteValue)
    _registry.register("StartsWith", StartsWith)
    _registry.register("EndsWith", EndsWith)
    _registry.register("StartsWithAny", StartsWithAny)
    _registry.register("EndsWithAny", EndsWithAny)
    _registry.register("Contains", Contains)
    _registry.register("CurrentYear", CurrentYear)
    _registry.register("IntervalCatV0", IntervalCatV0)
    _registry.register("LowerCase", LowerCase)
    _registry.register("IsSubStringOf", IsSubStringOf)
    _registry.register("GetChar", GetChar)

    return _registry


__all__ = [
    "registry",
    "BaseNode",
    "dynamic_nodes_registry",
    "BaseDynamicNode",
    "BaseConnector",
]



import pandas as pd
import pydantic

from retrack.nodes.base import BaseNode, InputConnectionModel, OutputConnectionModel

################################################
# And Or Inputs and Outputs
################################################


class AndOrInputsModel(pydantic.BaseModel):
    input_bool_0: InputConnectionModel
    input_bool_1: InputConnectionModel


class AndOrOutputsModel(pydantic.BaseModel):
    output_bool: OutputConnectionModel


################################################
# Not Inputs and Outputs
################################################


class NotInputsModel(pydantic.BaseModel):
    input_bool: InputConnectionModel


class NotOutputsModel(pydantic.BaseModel):
    output_bool: OutputConnectionModel


################################################
# And Or Nodes
################################################


class And(BaseNode):
    inputs: AndOrInputsModel
    outputs: AndOrOutputsModel

    def run(self, input_bool_0: pd.Series, input_bool_1: pd.Series) -> pd.Series:
        return {"output_bool": input_bool_0 & input_bool_1}


class Or(BaseNode):
    inputs: AndOrInputsModel
    outputs: AndOrOutputsModel

    def run(self, input_bool_0: pd.Series, input_bool_1: pd.Series) -> pd.Series:
        return {"output_bool": input_bool_0 | input_bool_1}


################################################
# Not Nodes
################################################


class Not(BaseNode):
    inputs: NotInputsModel
    outputs: NotOutputsModel

    def run(self, input_bool: pd.Series) -> pd.Series:
        return {"output_bool": ~input_bool}



import enum
import typing

import pydantic

###############################################################
# Node Kind
###############################################################


class NodeKind(str, enum.Enum):
    INPUT = "input"
    CONSTANT = "constant"
    OUTPUT = "output"
    FILTER = "filter"
    CONNECTOR = "connector"
    START = "start"
    OTHER = "other"
    FLOW = "flow"


###############################################################
# Node Memory Types
###############################################################


class NodeMemoryType(str, enum.Enum):
    STATE = "state"
    FILTER = "filter"
    CONSTANT = "constant"


###############################################################
# Connection Models
###############################################################


def cast_int_to_str(v: typing.Any, info: pydantic.ValidationInfo) -> str:
    return str(v)


CastedToStringType = typing.Annotated[
    typing.Any, pydantic.BeforeValidator(cast_int_to_str)
]


def cast_empty_string_to_none(v: str, info: pydantic.ValidationInfo) -> typing.Any:
    if v == "":
        return None
    return v


OptionalCastedToNoneStringType = typing.Annotated[
    typing.Optional[str], pydantic.BeforeValidator(cast_empty_string_to_none)
]


class OutputConnectionItemModel(pydantic.BaseModel):
    node: CastedToStringType
    input: str = pydantic.Field(alias="input")


class InputConnectionItemModel(pydantic.BaseModel):
    node: CastedToStringType
    output: str = pydantic.Field(alias="output")


class OutputConnectionModel(pydantic.BaseModel):
    connections: typing.List[OutputConnectionItemModel]


class InputConnectionModel(pydantic.BaseModel):
    connections: typing.List[InputConnectionItemModel]


###############################################################
# Base Node
###############################################################


class BaseNode(pydantic.BaseModel):
    id: CastedToStringType
    name: str
    inputs: typing.Optional[typing.Dict[CastedToStringType, InputConnectionModel]] = (
        None
    )
    outputs: typing.Optional[typing.Dict[CastedToStringType, OutputConnectionModel]] = (
        None
    )

    def run(self, **kwargs) -> typing.Dict[str, typing.Any]:
        return {}

    def kind(self) -> NodeKind:
        return NodeKind.OTHER

    def memory_type(self) -> NodeMemoryType:
        return NodeMemoryType.STATE

    def generate_input_nodes(self) -> typing.List["BaseNode"]:
        return []



import typing

import pandas as pd
import pandera
import pydantic

from retrack.nodes.base import BaseNode, NodeKind


class StrFieldValidator:
    def __init__(self, default: typing.Optional[typing.Any] = None):
        self.default = default

    def __call__(self, value: typing.Any) -> typing.Any:
        if value in [None, "None", "", "null"]:
            if self.default is None:
                raise ValueError("value cannot be None")
            else:
                return self.default

        return str(value) if not isinstance(value, str) else value


class RequestManager:
    def __init__(self, inputs: typing.List[BaseNode]):
        self._model = None
        self._dataframe_model = None
        self.inputs = inputs

    @property
    def inputs(self) -> typing.List[BaseNode]:
        return self._inputs

    @property
    def input_names(self) -> typing.List[str]:
        return [input.data.name for input in self.inputs]

    @inputs.setter
    def inputs(self, inputs: typing.List[BaseNode]):
        if not isinstance(inputs, list):
            raise TypeError(f"inputs must be a list, not {type(inputs)}")

        formated_inputs = {}

        for i in range(len(inputs)):
            if not isinstance(inputs[i], BaseNode):
                raise TypeError(
                    f"inputs[{i}] must be an InputModel, not {type(inputs[i])}"
                )

            if (
                inputs[i].kind() != NodeKind.INPUT
                and inputs[i].kind() != NodeKind.CONNECTOR
            ):
                raise TypeError(
                    f"inputs[{i}] must be an InputModel, not {type(inputs[i])}"
                )

            if inputs[i].data.name not in formated_inputs:
                formated_inputs[inputs[i].data.name] = inputs[i]
            elif inputs[i].data.default is not None:
                formated_inputs[inputs[i].data.name] = inputs[i]

        self._inputs = formated_inputs.values()

        if len(self.inputs) > 0:
            self._model = self.__create_model()
            self._dataframe_model = self.__create_dataframe_model()
        else:
            self._model = None
            self._dataframe_model = None

    @property
    def model(self) -> typing.Type[pydantic.BaseModel]:
        return self._model

    @property
    def dataframe_model(self) -> pandera.DataFrameSchema:
        return self._dataframe_model

    def __create_model(
        self, model_name: str = "RequestModel"
    ) -> typing.Type[pydantic.BaseModel]:
        """Create a pydantic model from the RequestManager's inputs

        Args:
            model_name (str, optional): The name of the model. Defaults to "RequestModel".

        Returns:
            typing.Type[pydantic.BaseModel]: The pydantic model
        """
        fields = {}
        for input_field in self.inputs:
            fields[input_field.data.name] = (
                typing.Annotated[
                    str if input_field.data.default is None else typing.Optional[str],
                    pydantic.BeforeValidator(
                        StrFieldValidator(input_field.data.default)
                    ),
                ],
                pydantic.Field(
                    default=Ellipsis
                    if input_field.data.default is None
                    else input_field.data.default,
                    json_schema_extra={
                        "optional": input_field.data.default is not None
                    },
                    validate_default=False,
                ),
            )

        return pydantic.create_model(
            model_name,
            **fields,
            __config__=pydantic.ConfigDict(
                from_attributes=True, use_enum_values=True, protected_namespaces={}
            ),
        )

    def __create_dataframe_model(self) -> pandera.DataFrameSchema:
        """Create a pydantic model from the RequestManager's inputs"""
        fields = {}
        for input_field in self.inputs:
            fields[input_field.data.name] = pandera.Column(
                str,
                nullable=input_field.data.default is not None,
                coerce=True,
                default=input_field.data.default,
            )

        return pandera.DataFrameSchema(
            fields,
            index=pandera.Index(int),
            # strict=True,
            coerce=True,
        )

    def validate(
        self,
        payload: pd.DataFrame,
    ) -> pd.DataFrame:
        """Validate the payload against the RequestManager's model

        Args:
            payload (pandas.DataFrame): The payload to validate

        Raises:
            ValueError: If the RequestManager has no model

        Returns:
            pd.DataFrame: The validated payload
        """
        if self.model is None:
            raise ValueError("No inputs found")

        if not isinstance(payload, pd.DataFrame):
            raise TypeError(f"payload must be a pandas.DataFrame, not {type(payload)}")

        return self.dataframe_model.validate(payload)
